#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'

import asyncio
import collections
import json
import pickle

from aiopyramid.websocket.config.gunicorn import WebsocketMapper
from aiopyramid.websocket.view import WebsocketConnectionView
from pyramid.view import view_config

from pyams_cache.cache import get_cache_handler
from pyams_cache.interfaces import IAioCacheHandler
from pyams_notify_ws.subscribe import users, users_lock


queue = collections.deque(maxlen=100)
queue_key = b'PyAMS:notify:messages_queue'
queue_lock = asyncio.Lock()

cache_handler = None


def init_cache_client(server):
    """Initialize cache handler"""
    global cache_handler
    cache_handler = get_cache_handler(server, IAioCacheHandler)


@view_config(route_name='notify', mapper=WebsocketMapper)
class WebsocketNotifyView(WebsocketConnectionView):
    """Websocket notify view"""

    async def on_open(self):
        pass

    async def on_message(self, message):
        try:
            message = json.loads(message)
        except ValueError:
            pass
        else:
            action = message.get('action')
            if action == 'notify':
                # dispatch notification to subscribers
                source = message.get('source', {}).get('id')
                target = message.get('target', {}).get('principals', ())
                clean_message = message.copy()
                clean_message.pop('target')
                json_message = json.dumps(clean_message)
                async with users_lock:
                    for connection, subscription in users.items():
                        # don't send notification to emitter
                        if source and (source == subscription.principal['id']):
                            continue
                        if subscription.filter_target(target):
                            await connection.send(json_message)
                # store message in cached queue
                if cache_handler is not None:
                    async with queue_lock:
                        mem_queue = await cache_handler.get(queue_key)
                        if mem_queue is None:
                            mem_queue = queue
                        else:
                            mem_queue = pickle.loads(mem_queue)
                        mem_queue.append(message)
                        await cache_handler.set(queue_key, pickle.dumps(mem_queue))

    async def on_close(self):
        pass
